local roster = AceLibrary("Roster-2.1")
local aura = AceLibrary("SpecialEvents-Aura-2.0")

local lastDkpCacheUpdate = {}
local clientDkpCache = {}

local counterRunning = nil

-----------------------------------------------------------------------
-- Locale
-----------------------------------------------------------------------
local L = setmetatable({}, {__index =
	function(self, key)
		self[key] = key
		return key
	end
})

local flasks = {
	["Flask of Arcane Fortification"] = true,
	["Chromatic Resistance"] = true,
	["Distilled Wisdom"] = true,
	["Flask of Fortification"] = true,
	["Flask of Mighty Restoration"] = true,
	["Flask of Relentless Assault"] = true,
	["Flask of Shadow Fortification"] = true,
	["Supreme Power"] = true,
	["Flask of the Titans"] = true,
	["Supreme Power of Shattrath"] = true,
	["Fortification of Shattrath"] = true,
	["Mighty Restoration of Shattrath"] = true,
	["Relentless Assault of Shattrath"] = true,
	["Unstable Flask of the Bandit"] = true,
	["Unstable Flask of the Beast"] = true,
	["Unstable Flask of the Elder"] = true,
	["Unstable Flask of the Sorcerer"] = true,
	["Unstable Flask of the Soldier"] = true,
	["Unstable Flask of the Physician"] = true,
}

local elixirs = {
	["Healing Power"] = true,
	["Elixir of Draenic Wisdom"] = true,
	["Elixir of Brute Force"] = true,
	["Fel Strength Elixir"] = true,
	["Empowerment"] = true,
	["Greater Mana Regeneration"] = true,
	["Major Shadow Power"] = true,
	["Major Armor"] = true,
	["Elixir of Ironskin"] = true,
	["Major Agility"] = true,
	["Earthen Elixir"] = true,
	["Major Frost Power"] = true,
	["Elixir of Mastery"] = true,
	["Elixir of Major Fortitude"] = true,
	["Major Strength"] = true,
	["Adept's Elixir"] = true,
	["Onslaught Elixir"] = true,
	["Greater Arcane Elixir"] = true,
	["Elixir of the Mongoose"] = true,
	["Greater Firepower"] = true,
	["Shadow Power"] = true,
	["Elixir of the Giants"] = true,
	["Arcane Elixir"] = true,
}

local lootDialogOpen = nil
local lootQueue = {}

local halfHour = 60 * 30
local oneHour = halfHour * 2

local paused = nil
local counterRunning = nil

local db = nil

local _G = getfenv(0)

Symphony = BananaDKP:NewModule("Symphony")

Symphony.defaultDatabase = {
	databaseVersion = 2, -- Database version.
	hours = {}, -- Key: Player, Value: Hours of raiding, autoincremented every 30min.
	items = {}, -- Key: Player, Value: Table of item ids looted.
	lastUpdate = {}, -- Key: Player, Value: Time of last update for this player.
	dkp = {},
	hourTickerSavePoint = 0,
}

Symphony.options = {
	autoHeader = {
		type = "header",
		name = L["Automatic"],
		order = 100,
	},
	start = {
		type = "execute",
		name = L["Start"],
		desc = L["Start the automatic hour counter."],
		func = function() Symphony:StartCounter() end,
		order = 101,
		disabled = function() return not UnitInRaid("player") or counterRunning end,
	},
	pause = {
		type = "toggle",
		name = L["Pause"],
		desc = L["Pause the automatic hour counter."],
		get = function() return paused and true or false end,
		set = function(v) Symphony:PauseCounter(v) end,
		order = 102,
		disabled = function() return not counterRunning end,
	},
	stop = {
		type = "execute",
		name = L["Stop"],
		desc = L["Stop the automatic hour counter."],
		func = function() Symphony:StopCounter() end,
		disabled = function() return not counterRunning end,
		order = 103,
	},
	spacer = {
		type = "header",
		name = " ",
		order = 150,
	},
	manualHeader = {
		type = "header",
		name = L["Manual"],
		order = 200,
	},
	plusone = {
		type = "text",
		name = L["Increment"],
		desc = L["Increment the hour counter for a player or the entire raid+sitout list (*) by one."],
		usage = L["<player or *>"],
		get = false,
		set = function(player) Symphony:IncrementHourCounter(player) end,
		order = 201,
	},
	minusone = {
		type = "text",
		name = L["Decrement"],
		desc = L["Decrement the hour counter for a player or the entire raid+sitout list (*) by one."],
		usage = L["<player or *>"],
		get = false,
		set = function(player) Symphony:DecrementHourCounter(player) end,
		order = 202,
	},
}

function Symphony:OnEnable()
	local popup = _G.StaticPopupDialogs
	if type(popup) ~= "table" then
		popup = {}
	end

	if type(popup["SymphonyItemQuery"]) ~= "table" then
		popup["SymphonyItemQuery"] = {
			button1 = L["Add"],
			button2 = L["Ignore"],
			sound = "AuctionWindowClose",
			whileDead = 1,
			timeout = 0,
			OnAccept = function(data)
				Symphony:AddItemToItemTable(unpack(data))
			end,
			OnCancel = function()
				Symphony:IgnoreItem()
			end,
		}
	end

	db = self.db.profile

	self.core:RegisterWhisperHandler("items", function(event, player)
		local mode = BananaDKP:GetMode()
		if mode == 2 then
			local host = BananaDKP:GetHost()
			if not host then return end
			return "I'm not the DKP host, whisper " .. host .. " instead."
		end
		player = player:lower()
		if db.items[player] then
			local items = {}
			for i, v in ipairs(db.items[player]) do
				local itemId = select(3, v:find("item:(%d+):"))
				local name, link = GetItemInfo(itemId)
				if link then
					table.insert(items, link)
				elseif name then
					table.insert(items, name)
				else
					table.insert(items, (select(3, v:find("%[(.*)%]"))))
				end
			end
			return { string.format(L["Items looted by %s:"], player), items, string.format(L["(%d items)"], #items) }
		else
			return string.format(L["No items looted by %s"], player)
		end
	end, L["bdkp items <player>; list of the items the player has looted"])

	self:RegisterEvent("RosterLib_RosterUpdated")

	self:RegisterEvent("BananaDKP_HostChanged")
	self:RegisterEvent("BananaDKP_ModeChanged")

	self:RegisterEvent("CHAT_MSG_ADDON")
end

local function addonMsg(prefix, message, distribution, target)
	if ChatThrottleLib then
		ChatThrottleLib:SendAddonMessage("BULK", prefix, message, distribution, target)
	else
		SendAddonMessage(prefix, message, distribution, target)
	end
end

local grouped = nil
function Symphony:RosterLib_RosterUpdated()
	if not grouped and GetNumRaidMembers() > 0 then
		grouped = true
	elseif grouped and GetNumRaidMembers() == 0 then
		self:StopCounter()
		grouped = nil
	end
end

function Symphony:BananaDKP_ModeChanged(mode)
	if counterRunning then
		self:StopCounter()
	end
	self:TriggerEvent("BananaDKP_DKPUpdated")
end

function Symphony:BananaDKP_HostChanged(host)
	if counterRunning then
		self:StopCounter()
	end
	self:TriggerEvent("BananaDKP_DKPUpdated")
end

local function addItem(player, itemLink)
	player = player:lower()
	if not db.items[player] then db.items[player] = {} end
	table.insert(db.items[player], itemLink)

	-- DKP -25%
	local dkp = db.dkp[player]
	local price = 0
	if not dkp then
		dkp = 0
	else
		price = dkp * 25 / 100
	end
	db.dkp[player] = dkp - price

	db.lastUpdate[player] = time()

	return price
end

local function incrementByOne(input)
	input = input:lower()
	if db.dkp[input] then
		db.dkp[input] = db.dkp[input] + 1
	else
		db.dkp[input] = 1
	end
end

local function decrementByOne(input)
	input = input:lower()
	if db.dkp[input] then
		db.dkp[input] = db.dkp[input] - 1
	end
end

function Symphony:CHAT_MSG_ADDON(prefix, message, distribution, sender)
	if not sender then return end
	if prefix == "BDKPI" then
		if roster.roster[sender] then return end -- Ignore anything from our own group.
		local player, itemLink = select(3, message:find("(%S+) (.*)"))
		if not player and itemLink then return end
		local price = addItem(player, itemLink)
		self.core:Print(L["%s says %s looted %s, %d has been deducted from his DKP."]:format(sender, player, itemLink, price or 0))
		self:TriggerEvent("BananaDKP_DKPUpdated")
	elseif prefix == "BDKPH" then
		if roster.roster[sender] then return end -- Ignore anything from our own group.

		local input, action = select(3, message:find("(%S+) (.*)"))
		if action == "+" then
			incrementByOne(input)
		elseif action == "-" then
			decrementByOne(input)
		elseif tonumber(action) then
			action = tonumber(action)
			input = input:lower()
			if not db.dkp[input] then
				db.dkp[input] = action
			else
				db.dkp[input] = db.dkp[input] + action
			end
			db.lastUpdate[input] = time()
		end
		self:TriggerEvent("BananaDKP_DKPUpdated")
	elseif prefix == "BDKP" then
		local player, dkp = select(3, message:find("(%S+) (.*)"))
		clientDkpCache[player:lower()] = tonumber(dkp)
		self:TriggerEvent("BananaDKP_DKPUpdated")
	elseif prefix == "BDKPR" then
		local mode = BananaDKP:GetMode()
		if mode == 1 then
			local dkp = self:GetDKP(message)
			addonMsg("BDKP", message .. " " .. tostring(dkp), "WHISPER", sender)
		end
	end
end

-----------------------------------------------------------------------
-- Ticker handling

-- Every minute we save our current timer, so that if we get disconnected or
-- something, we know where to start from next time.
function Symphony:RepeatedTimeDBStore()
	if not counterRunning then return end
	local timeLeft
	if paused then
		timeLeft = paused
	else
		timeLeft = select(2, self:IsEventScheduled("Ticker"))
	end
	db.hourTickerSavePoint = timeLeft
	self:ScheduleEvent("TickerStore", self.RepeatedTimeDBStore, 60, self)
end

local function startCounter()
	local time = 0
	if db.hourTickerSavePoint > 0 then
		time = db.hourTickerSavePoint
	else
		time = halfHour
	end
	Symphony:ScheduleEvent("Ticker", Symphony.IncrementHourCounter, time, Symphony)
	counterRunning = true
	Symphony:RepeatedTimeDBStore()
	local minutes = math.floor(time / 60)
	BananaDKP:Print(L["Automatic hour counting has started, the first increment will be in ~%d minutes."]:format(minutes))
end

function Symphony:StartCounter()
	assert(not counterRunning, "Counter is already running.")
	startCounter()
end

function Symphony:PauseCounter(input)
	assert(counterRunning, "Can't pause something that is not running.")
	if input then
		paused = select(2, self:IsEventScheduled("Ticker"))
		self:CancelScheduledEvent("Ticker")
		local minutes = math.floor(paused / 60)
		self.core:Print(string.format(L["Automatic hour counting paused, ~%d minutes to next update."], minutes))
	else
		self:ScheduleEvent("Ticker", self.IncrementHourCounter, paused, self)
		paused = nil
		self.core:Print(L["Automatic hour counting resumed."])
	end
end

function Symphony:StopCounter()
	if not counterRunning then return end
	local timeLeft = select(2, self:IsEventScheduled("Ticker"))
	if not timeLeft then timeLeft = paused end
	self:CancelScheduledEvent("Ticker")
	self:CancelScheduledEvent("TickerStore")
	counterRunning = nil
	timeLeft = math.floor(timeLeft / 60)
	self.core:Print(string.format(L["Automatic hour counting stopped, ~%d minutes to next update."], timeLeft or -1))
	paused = nil
	db.hourTickerSavePoint = 0
end

-----------------------------------------------------------------------
-- Item handling

function Symphony:ItemLooted(player, itemLink, itemRarity)
	local mode = BananaDKP:GetMode()
	local host = BananaDKP:GetHost()
	if mode == 2 and host and roster.roster[host] then return end

	if itemRarity < 4 then return end
	if lootDialogOpen then
		table.insert(lootQueue, {player, itemLink, itemRarity})
	else
		lootDialogOpen = true
		StaticPopupDialogs["SymphonyItemQuery"].text = string.format(L["%s looted %s, do you want to add it to his loot table?"], player, itemLink)
		local popup = StaticPopup_Show("SymphonyItemQuery")
		if popup then
			popup.data = {player, itemLink}
		end
	end
end

function Symphony:AddItemToItemTable(player, itemLink)
	lootDialogOpen = nil

	-- mode 1 = server, 2 = client
	local mode = BananaDKP:GetMode()
	if mode == 1 then
		local price = addItem(player, itemLink)
		self.core:Print(L["%s looted %s, %d has been deducted from his DKP."]:format(player, itemLink, price or 0))
	elseif mode == 2 then
		local host = BananaDKP:GetHost()
		if not host then return end
		addonMsg("BDKPI", player .. " " .. itemLink, "WHISPER", host)
		self.core:Print(L["%s looted %s, it has been added to his item table."]:format(player, itemLink))
	end

	if #lootQueue > 0 then
		local n = table.remove(lootQueue, 1)
		self:ItemLooted(n[1], n[2], n[3])
	else
		self:TriggerEvent("BananaDKP_DKPUpdated")
	end
end
function Symphony:IgnoreItem()
	lootDialogOpen = nil

	if #lootQueue > 0 then
		self:ItemLooted(unpack(table.remove(lootQueue, 1)))
	end
end

-----------------------------------------------------------------------
-- Hour handling

function Symphony:IncrementHourCounter(input)
	local mode = BananaDKP:GetMode()
	local host = BananaDKP:GetHost()

	if type(input) == "string" and not input:find("*") then
		input = input:lower()
		if mode == 1 then
			incrementByOne(input)
		elseif mode == 2 then
			if not host then return end
			addonMsg("BDKPH", input .. " +", "WHISPER", host)
		end
		self.core:Print(string.format(L["The hour counter has been incremented by 1 for %s."], input))
	else
		if type(input) == "string" and input ~= "*" then
			error("Can't increment the hour counter for "..input..".")
			return
		end
		local dkpTable = {}
		for n, u in pairs(roster.roster) do
			if u.class ~= "PET" then
				local multiplier = 1
				for name, index in aura:BuffIter(u.unitid) do
					if flasks[name] then
						multiplier = multiplier + 0.4
					elseif name == "Well Fed" then
						multiplier = multiplier + 0.1
					elseif elixirs[name] then
						multiplier = multiplier + 0.2
					end
				end
				local amount = 1 * multiplier
				local k = n:lower()
				dkpTable[k] = amount
			end
		end
		for k, v in pairs(BananaDKP.db.profile.sitouts) do
			local key = k:lower()
			dkpTable[key] = 1
		end
		if mode == 1 then
			for k, v in pairs(dkpTable) do
				if not db.dkp[k] then
					db.dkp[k] = v
				else
					db.dkp[k] = db.dkp[k] + v
				end
				db.lastUpdate[k] = time()
			end
		elseif mode == 2 then
			if not host then return end
			for k, v in pairs(dkpTable) do
				addonMsg("BDKPH", k .. " " .. tostring(v), "WHISPER", host)
			end
		end
		self.core:Print(L["The hour counter has been incremented by 1 for all players in the raid and on the sitout list."])
		if type(input) ~= "string" then
			self:ScheduleEvent("Ticker", self.IncrementHourCounter, oneHour, self)
		end
	end

	self:TriggerEvent("BananaDKP_DKPUpdated")
end

function Symphony:DecrementHourCounter(player)
	local mode = BananaDKP:GetMode()
	local host = BananaDKP:GetHost()

	if type(input) == "string" and not input:find("*") then
		input = input:lower()
		if mode == 1 then
			decrementByOne(input)
		elseif mode == 2 then
			if not host then return end
			addonMsg("BDKPH", input .. " -", "WHISPER", host)
		end
		self.core:Print(string.format(L["The hour counter has been decremented by 1 for %s."], input))
	else
		if type(input) == "string" and input ~= "*" then
			error("Can't decrement the hour counter for "..input..".")
			return
		end
		local dkpTable = {}
		for n, u in pairs(roster.roster) do
			if u.class ~= "PET" then
				dkpTable[n:lower()] = true
			end
		end
		for k, v in pairs(BananaDKP.db.profile.sitouts) do
			dbkTable[k:lower()] = true
		end
		if mode == 1 then
			for k, v in pairs(dkpTable) do
				if db.dkp[k] then
					db.dkp[k] = db.dkp[k] - 1
				end
				db.lastUpdate[k] = time()
			end
		elseif mode == 2 then
			if not host then return end
			for k, v in pairs(dkpTable) do
				addonMsg("BDKPH", k .. " -", "WHISPER", host)
			end
		end
		self.core:Print(L["The hour counter has been decremented by 1 for all players in the raid and on the sitout list."])
		if type(input) ~= "string" then
			self:ScheduleEvent("Ticker", self.IncrementHourCounter, oneHour, self)
		end
	end

	self:TriggerEvent("BananaDKP_DKPUpdated")
end

-----------------------------------------------------------------------
-- DKP handling

-- DKP Formula:
-- #hoursRaided / (1 + #itemsLooted)
--
-- Pretty simple, I know :P That's what my guild wanted, so that's what I did.

-- Thanks, Wobin.
local function round(n,shift)
	shift = 10^shift
	return math.floor((n*shift)+0.5)/shift
end

function Symphony:GetDKP(name)
	if type(name) ~= "string" then return nil end
	local mode = BananaDKP:GetMode()
	name = name:lower()

	if mode == 1 then
		if not db.dkp[name] then return nil end
		return db.dkp[name]
	elseif mode == 2 then
		if not lastDkpCacheUpdate[name] or (GetTime() > (lastDkpCacheUpdate[name] + 60)) then
			lastDkpCacheUpdate[name] = GetTime()
			local host = BananaDKP:GetHost()
			if not host then return end
			addonMsg("BDKPR", name, "WHISPER", host)
		end
		return clientDkpCache[name]
	end
end

function Symphony:GetLootedItems(name)
	return db.items[name:lower()]
end

